from importlib.util import module_from_spec, spec_from_file_location
import logging

import click

from .proxy import Proxy


log = logging.getLogger(__name__)


def configure_logging(verbosity):
    level = logging.ERROR - verbosity * 10
    logging.basicConfig(format='[%(levelname)s] :: %(message)s', level=level)


def load_script(path):
    spec = spec_from_file_location('script', path)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    log.info('Loaded script %s.', path)


@click.group()
@click.version_option(message='%(version)s')
def cli():
    """Padsniff: A Puzzle & Dragons HTTP Sniffer"""


@cli.command()
@click.option('--port', '-p', default=8080, help='Proxy service port.')
@click.option('--script', '-s', 'scripts', multiple=True, type=click.Path(exists=True),
              help='Load a Python script containing a callback. Can be supplied multiple times.')
@click.option('--verbose', '-v', 'verbosity', count=True,
              help='Increase logging level (default: ERROR). Can be supplied multiple times.')
def run(port, scripts, verbosity):
    """Run the proxy service."""
    configure_logging(verbosity)

    # load scripts before instantiating the proxy to allow the use of the `on` decorator
    for path in scripts:
        load_script(path)

    proxy = Proxy(host='0.0.0.0', port=port)
    proxy.run()



import logging

from mitmproxy.controller import handler as flow_handler
from mitmproxy.master import Master as FlowMaster
from mitmproxy.options import Options
from mitmproxy.proxy import ProxyConfig, ProxyServer

from .constants import ALL, GUNGHO_API_ENDPOINT, GUNGHO_USER_AGENT
from .structures import CaseInsensitiveDefaultDict
from .parallel import parallelize


log = logging.getLogger(__name__)


class BaseProxy(FlowMaster):


    def __init__(self, **options):
        opts = Options(**options)
        server = ProxyServer(ProxyConfig(opts))
        super().__init__(opts, server)


    def run(self):
        try:
            logging.info('Starting proxy on port %s.', self.options.listen_port)
            super().run()
        except KeyboardInterrupt:
            pass
        finally:
            logging.info('Shutting down.')
            self.shutdown()


class Proxy(BaseProxy):

    handlers = CaseInsensitiveDefaultDict(set)

    def __init__(self, host='0.0.0.0', port=8080):
        super().__init__(listen_host=host, listen_port=port, mode='transparent')
        self.handlers = type(self).handlers.copy()


    @flow_handler
    def response(self, flow):
        request = flow.request

        log.debug('Received response from %s request to %s.', request.method, request.pretty_url)

        if is_gungho(request):
            log.info('Captured %s request to %s. Forwarding flow to routing.', request.method, request.pretty_url)
            self.route(flow)


    def on(self, action, *, blocking=False):
        """
        Register a function to be called when a response is received from a request to `action`.

        The function will be called with the request and response. The response will not be received by the client
        until the decorated function returns.

        Read about possible actions on the padsniff wiki: https://bitbucket.org/necromanteion/padsniff/wiki/Home
        """
        def wrapper(func):
            self.handlers[action].add(func if blocking else parallelize(func))
            return func

        return wrapper


    def route(self, flow):
        """
        Route a flow's request and response to a handler function based on its action URL parameter.
        """
        request, response = flow.request, flow.response
        action = request.query.get('action')
        funcs = self.handlers[action] | self.handlers[ALL]

        for func in funcs:
            try:
                func(request, response)
            except:
                log.exception('Error while executing %s.', func.__name__)


def is_gungho(request):
    """Validate that `request` originated from a GungHo app."""
    return (request.headers.get('user-agent') == GUNGHO_USER_AGENT and
            request.path.startswith(GUNGHO_API_ENDPOINT))


def on(action, *, blocking=False, cls=Proxy):
    """
    Register a function to be called when a response is received from a request to `action`.

    This is a shortcut for `Proxy.on`, useful for registering functions before instantiating
    a `Proxy` object.
    """
    return cls.on(cls, action, blocking=blocking)



from copy import deepcopy
from threading import Thread
from weakref import WeakSet


class HandlerThread(Thread):
    """Thread subclass that keeps track of its instances."""

    instances = WeakSet()

    def __new__(cls, *args, **kwargs):
        inst = super().__new__(cls)
        cls.instances.add(inst)
        return inst


def parallelize(func, *, cls=HandlerThread):
    """
    Decorator to run `func` in parallel with the main thread/process.

    `cls` can be any callable that implements a threading.Thread-like
    signature and a start method, including multiprocessing.Process.

    Arguments passed to the decorated function are deep-copied to
    allow for thread-safe behavior.
    """
    qualname = getattr(func, '__qualname__', None)

    def run(*args, **kwargs):
        """Wraps execution of %s(*args, **kwargs) in a thread.""" % qualname
        t = cls(
            target=func,
            name=qualname,
            args=deepcopy(args),
            kwargs=deepcopy(kwargs),
        )
        t.start()

    # apply __wrapped__ attr to be consistent with functools.wraps
    run.__wrapped__ = getattr(func, '__wrapped__', func)

    return run



#!/usr/bin/env python3

from collections import Counter
import padsniff
import json


COUNTER = Counter()


@padsniff.on(action='clear_dungeon')
def log_plus_eggs(request, response):
    cards = json.loads(response.content.decode())['cards']

    for card in cards:
        COUNTER['hp'] += card[6]
        COUNTER['atk'] += card[7]
        COUNTER['rcv'] += card[8]

    COUNTER['runs'] += 1

    print(format_log(COUNTER))


def format_log(d):
    s = (
        'runs: {0[runs]:>3d} | '
        'hp: {0[hp]:>3d} | '
        'atk: {0[atk]:>3d} | '
        'rcv: {0[rcv]:>3d}'
    )

    return s.format(d)



from collections.abc import MutableMapping


class CaseInsensitiveDefaultDict(MutableMapping):
    """
    A mashup of a case-insensitive keyed dict and collections.defaultdict.

    Write an example for me!
    """


    @staticmethod
    def _transform(key):
        """Supports non-string keys."""
        try:
            return key.casefold()

        except AttributeError:
            return key


    def __init__(self, default_factory, data=None, **kwargs):
        self._dict = {}

        if default_factory is not None and not callable(default_factory):
            raise TypeError('default factory must be callable or None')

        self.default_factory = default_factory

        if data is None:
            data = {}

        self.update(data, **kwargs)


    def __setitem__(self, key, value):
        lookup_key = self._transform(key)
        self._dict[lookup_key] = (key, value)


    def __getitem__(self, key):
        lookup_key = self._transform(key)

        try:
            orig_key, value = self._dict[lookup_key]

        except KeyError:
            # call to __missing__ must be explicit because of overriding __getitem__
            value = self.__missing__(key)

        return value


    def __delitem__(self, key):
        lookup_key = self._transform(key)

        try:
            del self._dict[lookup_key]

        except KeyError:
            raise KeyError(key)


    def __iter__(self):
        return (orig_key for orig_key, value in self._dict.values())


    def __len__(self):
        return len(self._dict)


    def __contains__(self, key):
        # MutableMapping.__contains__ depends on __getitem__ raising a KeyError for missing keys
        # we have to override this behavior to prevent from always returning True unless default_factory is None
        return self._transform(key) in self._dict


    def __missing__(self, key):
        # emulate defaultdict's behavior is no default_factory is supplied
        if self.default_factory is None:
            raise KeyError(key)

        self[key] = self.default_factory()
        return self[key]


    def __repr__(self):
        rep = '{qualname}({default_factory}, {contents})'.format(
            qualname=self.__class__.__qualname__,
            default_factory=getattr(self.default_factory, '__qualname__', repr(self.default_factory)),
            contents=dict(self.items()),
        )

        return rep


    def copy(self):
        """Create a shallow copy of `self`."""
        cls = type(self)
        return cls(self.default_factory, self._dict.values())


    @classmethod
    def fromkeys(cls, default_factory, seq, value=None):
        """Create a new mapping with each key in `seq` set to `value`."""
        return cls(default_factory, dict.fromkeys(seq, value))


    def get(self, key, default=None):
        """
        Return the value associated with `key`, or `default` if `key` is not found.

        This method does not invoke the default factory.
        """
        lookup_key = self._transform(key)
        default_pair = (key, default)

        # _dict.get avoids __getitem__'s defaultdict-like behavior
        orig_key, value = self._dict.get(lookup_key, default_pair)

        return value


    def setdefault(self, key, default=None):
        """
        Return the value associated with `key`, setting it to `default` if `key` is not found.

        This method does not invoke the default factory.
        """
        lookup_key = self._transform(key)
        default_pair = (key, default)

        # _dict.setdefault avoids __getitems__'s defaultdict-like behavior
        orig_key, value = self._dict.setdefault(lookup_key, default_pair)

        return value



from setuptools import find_packages, setup

with open('requirements.txt') as f:
    install_requires = f.readlines()

with open('dev-requirements.txt') as f:
    tests_require = install_requires + f.readlines()

setup(
    name='padsniff',
    version='1.1.1',
    description='Command-line tool to sniff Puzzle & Dragons data.',
    long_description_markdown_filename='README.md',
    url='https://bitbucket.org/necromanteion/padsniff',
    author='Will Medlar',
    author_email='wmmedlar@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Games/Entertainment :: Puzzle Games',
    ],
    keywords=[
        'puzzle & dragons',
        'man-in-the-middle',
        'reverse engineering',
    ],
    install_requires=install_requires,
    setup_requires=[
        'pypandoc',
        'pytest-runner',
        'setuptools-markdown',
    ],
    tests_require=tests_require,
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'padsniff = padsniff.cli:cli'
        ]
    }
)
